import sys
import os
import logging
import traceback
import json
import uuid
import cloudpickle
from distributed import Future
from apscheduler.schedulers.background import BackgroundScheduler
from google.cloud import storage
from google.cloud import secretmanager
from dask.distributed import Client
import waterhealer as wh
import json
import os
import threading
import itertools
from datetime import datetime
import time
from typing import List
from . import streaming

logger = logging.getLogger()

__version__ = '0.0.33'

client = secretmanager.SecretManagerServiceClient()

try:
    secret = os.environ.get('kafka_bootstrap')
    payload = json.loads(
        client.access_secret_version(secret).payload.data.decode('utf-8')
    )
    kafka_bootstrap = payload.get('bootstrap')
except:
    raise Exception('please export a correct variable `kafka_bootstrap`.')

try:
    secret = os.environ.get('dask_bootstrap')
    payload = json.loads(
        client.access_secret_version(secret).payload.data.decode('utf-8')
    )
    dask_bootstrap = payload.get('bootstrap')

except:
    raise Exception('please export a correct variable `dask_bootstrap`.')

storage_client = storage.Client()
bucket_name = os.environ.get('storage_bootstrap', 'general-artifact')
bucket = storage_client.bucket(bucket_name)


def get_source(
    topics: List[str],
    consumer: str,
    get_dask: bool = True,
    boilerplates: List[str] = None,
    **kwargs,
):
    """
    Get Kafka source and Dask client.

    Parameters
    ----------
    topics : List[str]
        list of strings of topics want to subscribe.
    consumer : str
        consumer name.
    get_dask : bool, optional (default=True)
        connect to dask cluster.
    boilerplates : List[str], optional (default=None)
        list of strings of filename for import, example `['boilerplate_fave_payments.py']`

    Returns
    -------
    result: (kafka_source, dask_client)
    """

    source = wh.from_kafka(
        topics,
        {
            'bootstrap.servers': kafka_bootstrap,
            'group.id': consumer,
            'auto.offset.reset': kwargs.get('offset', 'earliest'),
            'max.poll.interval.ms': kwargs.get('maxpoll', 30_000_000),
        },
        maxage_memory = kwargs.get('maxage', 9_999_999),
    )
    if get_dask:

        def reload_package(package):
            import sys
            import importlib

            try:

                importlib.reload(sys.modules[package])
            except:
                pass

            return list(sys.modules.keys())

        client = Client(dask_bootstrap, loop = source.loop)
        if boilerplates:
            for boilerplate in boilerplates:
                client.upload_file(boilerplate)
                client.run(reload_package, boilerplate.replace('.py', ''))
    else:
        client = None
    return source, client


def get_args(client, key):
    def _get_args(key, dask_scheduler = None):
        ts = dask_scheduler.tasks.get(key)
        if ts:
            args = ts.run_spec
        else:
            args = None
        return {'task': args}

    s = client.run_on_scheduler(_get_args, key)
    if 'args' in s['task']:
        args = cloudpickle.loads(s['task']['args'])
    else:
        args = None

    if args:
        if isinstance(args, tuple) or isinstance(args, list):
            if isinstance(args[0], tuple) or isinstance(args[0], list):
                if isinstance(args[0][0], tuple) or isinstance(
                    args[0][0], list
                ):
                    data = args[0][0]
                else:
                    data = args[0]
            else:
                data = args
        else:
            data = [args]
    return data


processed = {}


def logging_dask(client, expired: int = 1800, debug = False):
    """
    Parameters
    ----------
    client: object, (default=None)
        should be a dask client.
    expired: int, (default=1800)
        after expired time, logging dask will no longer check it.
    debug: bool, (default=False)
        if True, will print any logs related to logging dask.
    """
    global processed

    class Gather(threading.Thread):
        def __init__(self, key, debug):
            self.key = key
            self.future = Future(key)
            self.cache = {}
            self.debug = debug
            threading.Thread.__init__(self)

        def exit(self):
            print(f'{self.key} exited.')
            sys.exit()

        def run(self):
            while True:
                status = self.future._state.status

                if status == 'error':
                    func, args, kwargs, _ = client.get_futures_error(
                        self.future
                    )
                    func_name = f'{func.__module__}.{func.__name__}'

                    if isinstance(args, tuple) or isinstance(args, list):
                        if len(args[0]) == 2:
                            if 'dict' not in args[0][1]:
                                data = args[0]
                            else:
                                data = args
                        else:
                            data = args
                        args_data = [self.cache.get(future) for future in data]

                    else:
                        args_data = [args]

                    if len(args_data):

                        filename = f'{func_name}-{str(uuid.uuid4())}.json'
                        exception = traceback.format_tb(self.future.traceback())
                        data = {
                            'function': func_name,
                            'args': args_data,
                            'kwargs': kwargs,
                            'exception': exception,
                            'key': self.key,
                        }
                        with open(filename, 'w') as fopen:
                            json.dump(data, fopen)
                        bucket_filename = 'streaming-logs/' + filename
                        blob = bucket.blob(bucket_filename)
                        blob.upload_from_filename(filename)
                        message = {
                            'function': func_name,
                            'logs': f'{bucket_name}/{bucket_filename}',
                            'fulllogs': f'https://storage.cloud.google.com/{bucket_name}/{bucket_filename}?authuser=1&cloudshell=false',
                        }
                        logger.error(
                            f'error {sys.argv[0]}.{func_name}', extra = message
                        )
                        print(f'{self.key} error')
                        self.exit()

                if status == 'finished':
                    print(f'{self.key} finished')
                    self.exit()

                if status == 'pending':
                    try:
                        data = get_args(client, self.key)
                        fs = {}
                        for key in data:
                            if key not in self.cache:
                                fs[key] = Future(key)
                        if len(fs):
                            fs = client.gather(
                                fs, errors = 'skip', direct = True
                            )
                            for key, v in fs.items():
                                if v is not None:
                                    self.cache[key] = v

                    except Exception as e:
                        if self.debug:
                            print(traceback.format_exc())
                            print('Exception dask logging', self.key, e)

    start_time = datetime.now()

    while True:
        if (datetime.now() - start_time).seconds < expired:
            for key, v in client.futures.copy().items():
                if (
                    key not in processed
                    and 'dict' not in key
                    and 'str' not in key
                    and 'byte' not in key
                    and 'json_loads' not in key
                ):
                    print(key)
                    t = Gather(key)
                    t.start()
                    processed[key] = True
        else:
            time.sleep(600)
