import requests
import os
import json
from datetime import datetime, date, timedelta
from kafka import KafkaProducer
from kafka.partitioner import RoundRobinPartitioner
from google.cloud import secretmanager
import pandas as pd
import logging


# ----------------------------------------------------------------------
client = secretmanager.SecretManagerServiceClient()


# ----------------------------------------------------------------------
try:
    secret = os.environ.get('prometheus_bootstrap')
    payload = json.loads(
        client.access_secret_version(secret).payload.data.decode('utf-8')
    )
    prometheus_bootstrap = payload.get('bootstrap')
except:
    raise Exception('please export a correct variable `prometheus_bootstrap`.')

try:
    secret = os.environ.get('kafka_bootstrap')
    payload = json.loads(
        client.access_secret_version(secret).payload.data.decode('utf-8')
    )
    kafka_bootstrap = payload.get('bootstrap')
except:
    raise Exception('please export a correct variable `kafka_bootstrap`.')

kafka_producer = KafkaProducer(
    bootstrap_servers = kafka_bootstrap.split(','),
    api_version = (0, 10),
    partitioner = RoundRobinPartitioner(),
)


# ----------------------------------------------------------------------
def publish_kafka(row, topic):
    """
    Push a data to a kafka topic.
    Parameters
    ----------
    row : Dict
    topic : str
    Returns
    -------
    result : bool
    """
    try:
        value = json.dumps(row)
        value_bytes = bytes(value, encoding = 'utf-8')
        x = kafka_producer.send(topic, value = value_bytes)
        return True
    except Exception as ex:
        print('Exception in publishing message')
        print(str(ex))
        return False


# ----------------------------------------------------------------------
def push_to_kafka(row, topic):
    if 'payment_state' in row:
        if row['payment_state'] in ['Successful', 'Cancelled/Refunded']:
            publish_kafka(row, topic)

    return row


# ----------------------------------------------------------------------
def separate_uuid(row):
    """
    separate uuid key from a dictionary.
    Parameters
    ----------
    row : Dict
    Returns
    -------
    result : (uuid, row)
    """

    if not row:
        return (None, None)

    uuid = row.pop('uuid')

    return (uuid, row)


# ----------------------------------------------------------------------
def push_prometheus(row, name, logging: bool = True):
    """
    push to prometheus HTTP.
    Parameters
    ----------
    row : Dict
    name : str
        name of prometheus metric.
    logging : bool, optional (default=True)
        print value in row before exit.
    Returns
    -------
    result : bool
    """
    success = False
    if row['data']['status']:
        try:
            requests.get(f'{prometheus_bootstrap}{name}', timeout = 3)
            success = True
        except Exception as e:
            if logging:
                print(e)
    if logging:
        print(datetime.now(), row)
    return success


# ----------------------------------------------------------------------
def groupby_sort(rows):

    if not len(rows):
        return []

    df = pd.DataFrame(rows)
    return df.sort_values('updated_at').to_dict('records')


# ----------------------------------------------------------------------
# Used in failed streaming script only
def update_failed_uuid(rows):

    if not len(rows):
        return []

    return rows[1]
